.form-signin input[id="email"] {
    margin-bottom: -1px;
    min-width: 320px;
    border-bottom-right-radius: 0;
    border-bottom-left-radius: 0;
}

.form-signin input[id="password"] {
    margin-bottom: 10px;
    border-top-left-radius: 0;
    border-top-right-radius: 0;
}

body {
    background-color: #f7f7f7;
}
