"use strict";(self.webpackChunk_JUPYTERLAB_CORE_OUTPUT=self.webpackChunk_JUPYTERLAB_CORE_OUTPUT||[]).push([[4665],{94665:(e,t,r)=>{r.d(t,{O:()=>T,g:()=>i});var n=r(26169);const T=new n.Token("@jupyterlab/settingeditor:ISettingEditorTracker"),i=new n.Token("@jupyterlab/settingeditor:IJSONSettingEditorTracker")}}]);
//# sourceMappingURL=4665.aa19a41.js.map