from typing import NewType

TaskDescriptorId = NewType("TaskDescriptorId", str)
SubscriberId = NewType("SubscriberId", str)
TaskId = NewType("TaskId", str)
WorkerId = NewType("WorkerId", str)
ConfigId = NewType("ConfigId", str)
NodeId = NewType("NodeId", str)
InfraAppName = NewType("InfraAppName", str)
GraphName = NewType("GraphName", str)
Email = NewType("Email", str)
Password = NewType("Password", str)
