"use strict";(self.webpackChunk_JUPYTERLAB_CORE_OUTPUT=self.webpackChunk_JUPYTERLAB_CORE_OUTPUT||[]).push([[9733],{69733:(e,t,a)=>{a.r(t),a.d(t,{default:()=>r});var n=a(59674);const r=[{id:"@jupyterlab/ui-components-extension:labicon-manager",provides:n.ILabIconManager,autoStart:!0,activate:e=>Object.create(null)},{id:"@jupyterlab/settingeditor-extension:form-registry",provides:n.IFormComponentRegistry,autoStart:!0,activate:e=>new n.FormComponentRegistry}]}}]);
//# sourceMappingURL=9733.a3b2a7f.js.map