"use strict";(self.webpackChunk_jupyterlite_javascript_kernel_extension=self.webpackChunk_jupyterlite_javascript_kernel_extension||[]).push([[568],{568:(e,a,r)=>{r.r(a),r.d(a,{default:()=>i});var t=r(671),s=r(430),n=r(199);const i=[{id:"@jupyterlite/javascript-kernel-extension:kernel",autoStart:!0,requires:[s.IKernelSpecs],activate:(e,a)=>{const r=t.PageConfig.getBaseUrl();a.register({spec:{name:"javascript",display_name:"JavaScript (Web Worker)",language:"javascript",argv:[],resources:{"logo-32x32":"","logo-64x64":t.URLExt.join(r,"/kernelspecs/javascript.svg")}},create:async e=>new n.JavaScriptKernel(e)})}}]}}]);